@@ -12,7 +12,7 @@ String::License - detect source code license statements in a text string
 
 =head1 VERSION
 
-Version v0.0.6
+Version v0.0.7
 
 =head1 SYNOPSIS
 
@@ -35,12 +35,12 @@ and serializes them in a normalized format.
 
 =cut
 
-package String::License v0.0.6;
+package String::License v0.0.7;
 
-use Carp            qw(croak);
-use Log::Any        ();
-use Scalar::Util    qw(blessed);
-use List::SomeUtils qw(nsort_by uniq);
+use Carp         qw(croak);
+use Log::Any     ();
+use Scalar::Util qw(blessed);
+use List::Util   qw(uniq);
 use Array::IntSpan;
 use Regexp::Pattern::License 3.4.0;
 use Regexp::Pattern 0.2.12;
@@ -50,11 +50,13 @@ use String::License::Naming::SPDX;
 use namespace::clean;
 
 class Tag {
-	field $id :param;
+	field $name :param;
+	field $desc :param;
 	field $begin :param;
 	field $end :param;
 
-	method data () { return lc __CLASS__, $id, $begin, $end }
+	method name () { return $name }
+	method data () { return lc __CLASS__, $name, $desc, $begin, $end }
 }
 
 class Exception :isa(Tag) { }
@@ -62,9 +64,10 @@ class Exception :isa(Tag) { }
 class Flaw :isa(Tag) { }
 
 class Licensing {
-	field $id :param;
+	field $name :param;
+	field $desc :param;
 
-	method data () { return lc __CLASS__, $id }
+	method data () { return lc __CLASS__, $name, $desc }
 }
 
 class Fulltext :isa(Tag) { }
@@ -123,6 +126,8 @@ field $expr = '';
 
 field $coverage;
 field @loose_licensing;
+field %fulltext;
+field %grant;
 
 =back
 
@@ -152,34 +157,30 @@ ADJUST {
 
 =cut
 
-method note ( $name, $begin, $end )
+method note ( $name, $begin = undef, $end = undef )
 {
+	my $obj;
+
+	if ( ref($name) ) {
+		$obj = $name;
+		( undef, $name, undef, $begin, $end ) = $obj->data;
+	}
+
 	$log->tracef(
 		'noted %s: %d-%d "%s"',
 		$name, $begin, $end,
 		substr( $string, $begin, $end - $begin )
 	);
 
+	return $obj
+		if $obj;
+
 	return [ $begin, $end ];
 }
 
 method tag ($obj)
 {
-	my ( $name, $desc, $type, $id, $begin, $end );
-
-	( $type, $id, $begin, $end ) = $obj->data;
-
-	if ( ref($id) ) {
-		( $name, $desc ) = @$id;
-	}
-	else {
-		$name
-			= $self->best_value( $Regexp::Pattern::License::RE{$id}, 'name' );
-		$desc = $self->best_value(
-			$Regexp::Pattern::License::RE{$id},
-			qw(caption name)
-		);
-	}
+	my ( $type, $name, $desc, $begin, $end ) = $obj->data;
 
 	if ( $type eq 'licensing' ) {
 		push @loose_licensing, [ $type, $name, $desc ];
@@ -253,11 +254,14 @@ method best_value ( $hashref, @props )
 		for my $org ( $naming->list_schemes ) {
 			for ( keys %$hashref ) {
 				/$re_prop_attrs/;
-				next unless $+{prop} and $+{prop} eq $prop;
-				next unless $+{org}  and $+{org} eq $org;
-				next if $+{version};
-				next if $+{other};
-				next if $+{until_date};
+				next
+					if not defined $+{prop}
+					or $+{prop} ne $prop
+					or not defined $+{org}
+					or $+{org} ne $org
+					or defined $+{version}
+					or defined $+{other}
+					or defined $+{until_date};
 
 				$value = $hashref->{$_};
 				last PROPERTY;
@@ -269,11 +273,27 @@ method best_value ( $hashref, @props )
 	return $value;
 }
 
+method name_and_desc ($id)
+{
+	my ( $ref, %result );
+
+	$ref = $Regexp::Pattern::License::RE{$id};
+	$result{name} = $self->best_value( $ref, 'name' ) || $id;
+	$result{desc}
+		= $self->best_value( $ref, 'caption' ) || $ref->{name} || $id;
+
+	return \%result;
+}
+
+my $license_contains_license_re
+	= qr/^license:contains:license:([a-z][a-z0-9_]*)/;
 my $type_re
 	= qr/^type:([a-z][a-z0-9_]*)(?::([a-z][a-z0-9_]*))?(?::([a-z][a-z0-9_]*))?/;
 
 our %RE;
-my ( %L, @RE_EXCEPTION, @RE_LICENSE, @RE_NAME );
+my (%L, @EXCEPTIONS, @LICENSES, @NAMES, @USAGE, @SINGLEVERSION, @VERSIONED,
+	@UNVERSIONED, @COMBO, @GROUP
+);
 
 method init_licensepatterns ()
 {
@@ -336,12 +356,7 @@ method init_licensepatterns ()
 			-prefix => 'LOCAL_TRAIT_',
 		),
 		'License::version' => (
-			capture => 'numbered',
-			subject => 'trait',
-			-prefix => 'LOCAL_TRAIT_KEEP_',
-		),
-		'License::version_numberstring' => (
-			capture => 'numbered',
+			capture => 'named',
 			subject => 'trait',
 			-prefix => 'LOCAL_TRAIT_KEEP_',
 		),
@@ -382,16 +397,17 @@ method init_licensepatterns ()
 		),
 	);
 
-	@RE_EXCEPTION = sort map {/^EXCEPTION_(.*)/} keys(%RE);
-	@RE_LICENSE   = sort map {/^LICENSE_(.*)/} keys(%RE);
-	@RE_NAME      = sort map {/^NAME_(.*)/} keys(%RE);
-
-	foreach my $key ( grep {/^[a-z]/} keys(%Regexp::Pattern::License::RE) ) {
+	my @license_containers;
+	for my $key ( grep {/^[a-z]/} keys %Regexp::Pattern::License::RE ) {
 		my $val = $Regexp::Pattern::License::RE{$key};
-		$L{name}{$key} = $self->best_value( $val, 'name' ) || $key;
-		$L{caption}{$key}
-			= $self->best_value( $val, 'caption' ) || $val->{name} || $key;
-		foreach ( @{ $val->{tags} } ) {
+
+		( $L{name}{$key}, $L{caption}{$key} )
+			= @{ $self->name_and_desc($key) }{ 'name', 'desc' };
+		for ( @{ $val->{tags} } ) {
+			if (/$license_contains_license_re/) {
+				$L{contained_licenses}{$key}{$1} = undef;
+				push @license_containers, $key;
+			}
 			/$type_re/ or next;
 			$L{type}{$1}{$key} = 1;
 			if ( $2 and $1 eq 'singleversion' ) {
@@ -412,24 +428,54 @@ method init_licensepatterns ()
 			}
 		}
 	}
+	for my $container (@license_containers) {
+		for my $contained ( keys %{ $L{contained_licenses}{$container} } ) {
+			if ( exists $L{contained_licenses}->{$contained} ) {
+				$L{contained_licenses}{$container}{$_} = undef
+					for keys %{ $L{contained_licenses}{$contained} };
+			}
+		}
+	}
+
+	# list by amount of contained licenses (fewest first),
+	# then key length (longest first), then alphabetically
+	@LICENSES = sort {
+		keys %{ $L{contained_licenses}{$b} } <=>
+			keys %{ $L{contained_licenses}{$a} }
+			|| length($b) <=> length($a)
+			|| $a cmp $b
+		}
+		map {/^LICENSE_(.*)/} keys %RE;
+
+	# list by key length (longest first), then alphabetically
+	@NAMES = sort { length($b) <=> length($a) || $a cmp $b }
+		map {/^NAME_(.*)/} keys %RE;
+
+	# list alphabetically
+	@EXCEPTIONS = sort map {/^EXCEPTION_(.*)/} keys %RE;
+
+	@USAGE = grep { exists $L{type}{usage}{$_} } @LICENSES;
+	@SINGLEVERSION
+		= grep { exists $L{type}{singleversion}{$_} } @LICENSES;
+	@VERSIONED = grep { exists $L{type}{versioned}{$_} } @LICENSES;
+	@UNVERSIONED
+		= grep { exists $L{type}{unversioned}{$_} } @LICENSES;
+	@COMBO = grep { exists $L{type}{combo}{$_} } @LICENSES;
+	@GROUP = grep { exists $L{type}{group}{$_} } @LICENSES;
 
 	# FIXME: drop when perl doesn't mysteriously  freak out over it
-	foreach (qw(any_of)) {
-		$L{re_trait}{$_} = '';
-	}
+	$L{re_trait}{any_of} = '';
 
 	#<<<  do not let perltidy touch this (keep long regex on one line)
 	$L{multi_1} = qr/$RE{LOCAL_TRAIT_licensed_under}$RE{LOCAL_TRAIT_any_of}(?:[^.]|\.\S)*$RE{LOCAL_NAME_lgpl}$RE{LOCAL_TRAIT_KEEP_version}?/i;
 	$L{multi_2} = qr/$RE{LOCAL_TRAIT_licensed_under}$RE{LOCAL_TRAIT_any_of}(?:[^.]|\.\S)*$RE{LOCAL_NAME_gpl}$RE{LOCAL_TRAIT_KEEP_version}?/i;
-	$L{lgpl_5} = qr/$RE{LOCAL_TRAIT_licensed_under}$RE{LOCAL_NAME_lgpl}(?:$RE{LOCAL_TRAIT_by_fsf})?[,;:]?(?: either)? ?$RE{LOCAL_TRAIT_KEEP_version_numberstring},? $RE{LOCAL_TRAIT_or_at_option} $RE{LOCAL_TRAIT_KEEP_version_numberstring}/i;
+	$L{lgpl_5} = qr/$RE{LOCAL_TRAIT_licensed_under}$RE{LOCAL_NAME_lgpl}(?:$RE{LOCAL_TRAIT_by_fsf})?$RE{LOCAL_TRAIT_KEEP_version}(?:,? ?$RE{LOCAL_TRAIT_or_at_option} $RE{LOCAL_TRAIT_KEEP_version})?/i;
 	$L{gpl_7} = qr/either $RE{LOCAL_NAME_gpl}$RE{LOCAL_TRAIT_KEEP_version}?(?: \((?:the )?"?GPL"?\))?, or $RE{LOCAL_NAME_lgpl}$RE{LOCAL_TRAIT_KEEP_version}?/i;
 	$L{apache_1} = qr/$RE{LOCAL_NAME_apache}$RE{LOCAL_TRAIT_KEEP_version}?(?:(?: or)? [^ ,]*?apache[^ ,]*| \([^(),]+\))*,? or $RE{LOCAL_NAME_gpl}$RE{LOCAL_TRAIT_KEEP_version}?/i;
-	$L{apache_2} = qr/$RE{LOCAL_NAME_apache}$RE{LOCAL_TRAIT_KEEP_version}?(?:(?: or)? [^ ,]*?apache[^ ,]*| \([^(),]\))*,? or(?: the)? bsd(?:[ -](\d)-clause)?\b/i;
+	$L{apache_2} = qr/$RE{LOCAL_NAME_apache}$RE{LOCAL_TRAIT_KEEP_version}?(?:(?: or)? [^ ,]*?apache[^ ,]*| \([^(),]\))*,? or(?: the)? bsd(?P<version_bsd>[ -](\d)-clause)?\b/i;
 	$L{apache_4} = qr/$RE{LOCAL_NAME_apache}$RE{LOCAL_TRAIT_KEEP_version}?(?:(?: or)? [^ ,]*?apache[^ ,]*| \([^(),]\))*,? or $RE{LOCAL_NAME_mit}\b/i;
 	$L{fsful} = qr/This (\w+)(?: (?:file|script))? is free software; $RE{LOCAL_TRAIT_fsf_unlimited}/i;
 	$L{fsfullr} = qr/This (\w+)(?: (?:file|script))?  is free software; $RE{LOCAL_TRAIT_fsf_unlimited_retention}/i;
-	$L{trailing_space} = qr/\s+$/;
-	$L{LEFTANCHOR_version_of} = qr/^ of /;
 	#>>>
 }
 
@@ -484,13 +530,6 @@ method resolve ()
 {
 	$self->init_licensepatterns;
 
-	my @L_type_usage         = sort keys %{ $L{type}{usage} };
-	my @L_type_singleversion = sort keys %{ $L{type}{singleversion} };
-	my @L_type_versioned     = sort keys %{ $L{type}{versioned} };
-	my @L_type_unversioned   = sort keys %{ $L{type}{unversioned} };
-	my @L_type_combo         = sort keys %{ $L{type}{combo} };
-	my @L_type_group         = sort keys %{ $L{type}{group} };
-
 	my @spdx_gplver;
 
 	my @agpl = qw(agpl agpl_1 agpl_2 agpl_3);
@@ -498,13 +537,12 @@ method resolve ()
 	my @lgpl = qw(lgpl lgpl_2 lgpl_2_1 lgpl_3);
 
 	my %match;
-	my ( %grant, %license );
 
 	my $patterns2id = sub ( $stem, $version = undef ) {
 		return $stem
 			unless ($version);
-		$version =~ s/\.0$//g;
-		$version =~ s/\./_/g;
+		$version =~ tr/./_/;
+		$version =~ s/_0$//g;
 		return "${stem}_$version";
 	};
 	my $id2patterns = sub ($id) {
@@ -547,7 +585,8 @@ method resolve ()
 		$expr = join( ' or ', sort @spdx );
 		$self->tag(
 			Licensing->new(
-				id => [ $expr, $L{caption}{$legacy} || $legacy ]
+				name => $expr,
+				desc => $L{caption}{$legacy} || $legacy,
 			)
 		);
 	};
@@ -555,89 +594,84 @@ method resolve ()
 	# fulltext
 	$log->trace('scan for license fulltext');
 	my %pos_license;
-	foreach my $id (@RE_LICENSE) {
+	for my $id (@LICENSES) {
 		next unless ( $RE{"LICENSE_$id"} );
 		while ( $string =~ /$RE{"LICENSE_$id"}/g ) {
-			$pos_license{ $-[0] }{$id}
-				= $self->note( "license($id)", $-[0], $+[0] );
+			$pos_license{ $-[0] }{obj}{$id} = $self->note(
+				Fulltext->new(
+					%{ $self->name_and_desc($id) },
+					begin => $-[0], end => $+[0]
+				)
+			);
+			$pos_license{ $-[0] }{end}{$id} = $+[0];
 		}
 	}
-
-	foreach my $trait ( keys %{ $L{TRAITS_exception_prefix} } ) {
-
+	for my $trait ( keys %{ $L{TRAITS_exception_prefix} } ) {
 		next unless ( $string =~ /$RE{"TRAIT_$trait"}/ );
 		while ( $string =~ /$RE{"TRAIT_$trait"}/g ) {
 			next if $self->contains_tag( $-[0], $+[0] );
 			$self->note( $trait, $-[0], $+[0] );
 		}
 	}
-	foreach my $pos ( sort { $a <=> $b } keys %pos_license ) {
+	for my $pos ( sort { $a <=> $b } keys %pos_license ) {
 
 		# pick longest or most specific among matched license fulltexts
-		my @licenses = nsort_by { $pos_license{$pos}{$_}->[1] }
-		grep { $pos_license{$pos}{$_} ? $pos_license{$pos}{$_}->[1] : () } (
-			@L_type_group,
-			@L_type_combo,
-			@L_type_unversioned,
-			@L_type_versioned,
-			@L_type_singleversion,
-			@L_type_usage,
-		);
-		my $license = pop @licenses;
-		next unless ($license);
+		my ($longest)
+			= sort { $b <=> $a } values %{ $pos_license{$pos}{end} };
+		my @licenses = grep {
+			exists $pos_license{$pos}{end}{$_}
+				and $pos_license{$pos}{end}{$_} eq $longest
+		} @LICENSES;
+		my $id = $licenses[0];
 		next
-			if $self->contains_tag( $pos, $pos_license{$pos}{$license}->[1] );
-		$self->tag(
-			Fulltext->new(
-				id    => [ $license, $license ],   # TODO: try resolve caption
-				begin => $pos_license{$pos}{$license}->[0],
-				end   => $pos_license{$pos}{$license}->[1]
-			)
-		);
-		$license{$license} = 1;
+			if not $id
+			or $self->contains_tag( $pos, $pos_license{$pos}{end}{$id} );
+		$fulltext{$id} = $self->tag( $pos_license{$pos}{obj}{$id} );
 	}
 
 	# grant, stepwise
 	my @prefixes;
 	$log->trace('scan stepwise for license grant');
-	foreach my $trait ( keys %{ $L{TRAITS_grant_prefix} } ) {
-
+	for my $trait ( keys %{ $L{TRAITS_grant_prefix} } ) {
 		while ( $string =~ /$RE{"TRAIT_$trait"}/g ) {
 			next if $self->contains_tag( $-[0], $+[0] );
 			push @prefixes, $self->note( $trait, $-[0], $+[0] );
 		}
 	}
 	LICENSED_UNDER:
-	foreach my $licensed_under ( sort { $a->[1] <=> $b->[1] } @prefixes ) {
+	for my $licensed_under ( sort { $a->[1] <=> $b->[1] } @prefixes ) {
 		my $pos = $licensed_under->[1];
 
 		# possible grant names
 		my @grant_types = (
-			@L_type_combo,
-			@L_type_unversioned,
-			@L_type_versioned,
-			@L_type_singleversion,
-			@L_type_usage,
+			@COMBO,
+			@UNVERSIONED,
+			@VERSIONED,
+			@SINGLEVERSION,
+			@USAGE,
 		);
 
 		# optional grant version
-		my ( $version, $later );
+		my ( $version, $suffix );
 
 		# scan for prepended version
 		substr( $string, $pos ) =~ $RE{ANCHORLEFT_NAMED_version};
-		if ( $+{version_number} ) {
-			$self->note( 'version', $pos + $-[0], $pos + $+[0] );
+		if ( defined $+{version_number} ) {
+			$self->note( 'version',        $pos + $-[0], $pos + $+[0] );
+			$self->note( 'version_number', $pos + $-[1], $pos + $+[1] );
 			$version = $+{version_number};
-			if ( $+{version_later} ) {
-				$self->note( 'or_later', $pos + $-[2], $pos + $+[2] );
-				$later = $+{version_later};
+			if ( defined $+{version_later} ) {
+				$self->note( 'version_later', $pos + $-[2], $pos + $+[2] );
+				$suffix = '_or_later';
 			}
-			if (
-				substr( $string, $pos + $+[0] ) =~ $L{LEFTANCHOR_version_of} )
-			{
-				$self->note( 'version_of', $pos + $-[0], $pos + $+[0] );
+			if ( defined $+{version_only} ) {
+				$self->note( 'version_only', $pos + $-[4], $pos + $+[4] );
+				$suffix = '_only';
+			}
+			if ( defined $+{version_of} ) {
+				$self->note( 'version_of', $pos + $-[5], $pos + $+[5] );
 				$pos += $+[0];
-				@grant_types = @L_type_versioned;
+				@grant_types = @VERSIONED;
 			}
 			else {
 				$version = '';
@@ -645,7 +679,7 @@ method resolve ()
 		}
 
 		# scan for name
-		foreach my $id (@RE_NAME) {
+		for my $id (@NAMES) {
 			if ( substr( $string, $pos ) =~ $RE{"NAME_$id"} ) {
 				$match{$id}{name}{ $pos + $-[0] }
 					= $self->note( "name($id)", $pos + $-[0], $pos + $+[0] );
@@ -654,10 +688,15 @@ method resolve ()
 
 		# pick longest matched license name
 		# TODO: include all of most specific type when more are longest
-		my @names = nsort_by { $match{$_}{name}{$pos}->[1] }
-		grep { $match{$_} and $match{$_}{name} and $match{$_}{name}{$pos} }
-			@grant_types;
-		my $name = pop @names;
+		my @names = sort {
+			$match{$a}{name}{$pos}->[1] <=> $match{$b}{name}{$pos}->[1]
+			}
+			grep {
+					$match{$_}
+				and $match{$_}{name}
+				and $match{$_}{name}{$pos}
+			} @grant_types;
+		my $name = $names[-1];
 		if (    $name
 			and $match{$name}{name}{$pos}
 			and !$self->contains_tag( $pos, $match{$name}{name}{$pos}->[1] )
@@ -666,42 +705,68 @@ method resolve ()
 			my $pos_end = $pos = $match{$name}{name}{$pos}->[1];
 
 			# may include version
-			if ( !$version and grep { $_ eq $name } @L_type_versioned ) {
+			if ( !$version and grep { $_ eq $name } @VERSIONED ) {
 				substr( $string, $pos ) =~ $RE{ANCHORLEFT_NAMED_version};
-				if ( $+{version_number} ) {
-					$self->note( 'version', $pos + $-[0], $pos + $+[0] );
+				if ( defined $+{version_number} ) {
+					$self->note(
+						'version',
+						$pos + $-[0], $pos + $+[0]
+					);
+					$self->note(
+						'version_number',
+						$pos + $-[1], $pos + $+[1]
+					);
 					$version = $+{version_number};
-					$pos_end = $pos + $+[1];
 					if ( $+{version_later} ) {
-						$self->note( 'or_later', $pos + $-[2], $pos + $+[2] );
-						$later   = $+{version_later};
-						$pos_end = $pos + $+[2];
+						$self->note(
+							'version_later',
+							$pos + $-[2], $pos + $+[2]
+						);
+						$suffix = '_or_later';
 					}
+					if ( defined $+{version_only} ) {
+						$self->note(
+							'version_only',
+							$pos + $-[4], $pos + $+[4]
+						);
+						$suffix = '_only';
+					}
+					$pos_end = $pos + $+[0];
 				}
 			}
-			elsif ( !$version and grep { $_ eq $name } @L_type_singleversion )
-			{
+			elsif ( !$version and grep { $_ eq $name } @SINGLEVERSION ) {
 				substr( $string, $pos )
 					=~ $RE{ANCHORLEFT_NAMED_version_later};
-				if ( $+{version_later} ) {
-					$self->note( 'or_later', $pos + $-[1], $pos + $+[1] );
-					$later   = $+{version_later};
-					$pos_end = $pos + $+[1];
+				if ( defined $+{version_later} ) {
+					$self->note(
+						'version_later',
+						$pos + $-[1], $pos + $+[1]
+					);
+					$suffix  = '_or_later';
+					$pos_end = $pos + $+[0];
 				}
 			}
 			if ($version) {
-				$version =~ s/(?:\.0)+$//;
-				$version =~ s/\./_/g;
+				$version =~ tr/./_/;
+				$version =~ s/(?:_0)+$//;
 				$name .= "_$version";
 			}
-			if ($later) {
-				my $latername = "${name}_or_later";
-				$self->note( $latername, $licensed_under->[0], $pos_end );
-				$grant{$latername} = 1;
-				next LICENSED_UNDER if grep { $grant{$_} } @RE_NAME;
+			if ($suffix) {
+				my $latername = "$name$suffix";
+				$grant{$latername} = $self->note(
+					Grant->new(
+						%{ $self->name_and_desc($latername) },
+						begin => $licensed_under->[0], end => $pos_end
+					)
+				);
+				next LICENSED_UNDER if grep { $grant{$_} } @NAMES;
 			}
-			$self->note( "grant($name)", $licensed_under->[0], $pos_end );
-			$grant{$name} = 1;
+			$grant{$name} = $self->note(
+				Grant->new(
+					%{ $self->name_and_desc($name) },
+					begin => $licensed_under->[0], end => $pos_end
+				)
+			);
 		}
 	}
 
@@ -711,12 +776,16 @@ method resolve ()
 
 		# address in AGPL/GPL/LGPL
 		while ( $string =~ /$RE{TRAIT_addr_fsf}/g ) {
-			foreach (
+			for my $id (
 				qw(addr_fsf_franklin_steet addr_fsf_mass addr_fsf_temple))
 			{
-				if ( defined $+{$_} ) {
+				if ( defined $+{$id} ) {
 					$self->tag(
-						Flaw->new( id => $_, begin => $-[0], end => $+[0] ) );
+						Flaw->new(
+							%{ $self->name_and_desc($id) },
+							begin => $-[0], end => $+[0]
+						)
+					);
 				}
 			}
 		}
@@ -724,10 +793,14 @@ method resolve ()
 
 	# exceptions
 	# TODO: conditionally limit to AGPL/GPL/LGPL
-	foreach (@RE_EXCEPTION) {
-		if ( $string =~ $RE{"EXCEPTION_$_"} ) {
+	for my $id (@EXCEPTIONS) {
+		if ( $string =~ $RE{"EXCEPTION_$id"} ) {
 			$self->tag(
-				Exception->new( id => $_, begin => $-[0], end => $+[0] ) );
+				Exception->new(
+					%{ $self->name_and_desc($id) },
+					begin => $-[0], end => $+[0]
+				)
+			);
 		}
 	}
 
@@ -737,7 +810,11 @@ method resolve ()
 	# generated file
 	if ( $string =~ $RE{TRAIT_generated} ) {
 		$self->tag(
-			Flaw->new( id => 'generated', begin => $-[0], end => $+[0] ) );
+			Flaw->new(
+				%{ $self->name_and_desc('generated') },
+				begin => $-[0], end => $+[0]
+			)
+		);
 	}
 
 	# multi-licensing
@@ -766,19 +843,24 @@ method resolve ()
 	$gen_license->(@multilicenses) if (@multilicenses);
 
 	# LGPL
-	if ( grep { $match{$_}{name} } @lgpl ) {
+	for ( grep { $match{$_}{name} } @lgpl ) {
 		$log->trace('scan for LGPL fulltext/grant');
 
 		# LGPL, dual versions last
 		if ( $string =~ $L{lgpl_5} ) {
+
+			# TODO: simplify, and require Regexp::Pattern::License v3.11.0
+			my $v2 = $+{version_number_2} // $-{version_number}[1] || next;
+
 			$self->tag(
 				Grant->new(
-					id    => [ "LGPL-$1 or LGPL-$2", "LGPL (v$1 or v$2)" ],
+					name  => "LGPL-$+{version_number} or LGPL-$v2",
+					desc  => "LGPL (v$+{version_number} or v$v2)",
 					begin => $-[0], end => $+[0],
 				)
 			);
-			$match{ 'lgpl_' . $1 =~ tr/./_/r }{custom} = 1;
-			$match{ 'lgpl_' . $2 =~ tr/./_/r }{custom} = 1;
+			$match{ 'lgpl_' . $+{version_number} =~ tr/./_/r }{custom} = 1;
+			$match{ 'lgpl_' . $v2                =~ tr/./_/r }{custom} = 1;
 			$match{lgpl}{custom} = 1;
 		}
 	}
@@ -788,7 +870,10 @@ method resolve ()
 		$log->trace('scan for GPL or LGPL dual-license grant');
 		if ( $string =~ $L{gpl_7} ) {
 			$self->note( "grant(gpl#7)", $-[0], $+[0] );
-			$gen_license->( 'gpl', $1, $2, 'lgpl', $3, $4 );
+			$gen_license->(
+				'gpl',  $-{version_number}[0], $-{version_later}[0],
+				'lgpl', $-{version_number}[1], $-{version_later}[1],
+			);
 			$match{gpl}{custom}  = 1;
 			$match{lgpl}{custom} = 1;
 		}
@@ -800,23 +885,32 @@ method resolve ()
 		for ($string) {
 			if ( $string =~ $L{apache_1} ) {
 				$self->note( 'grant(apache#1)', $-[0], $+[0] );
-				$gen_license->( 'apache', $1, $2, 'gpl', $3, $4 );
-				$match{ $patterns2id->( 'apache', $1 ) }{custom} = 1;
+				$gen_license->(
+					'apache', $-{version_number}[0], $-{version_later}[0],
+					'gpl',    $-{version_number}[1], $-{version_later}[1],
+				);
+				$match{ $patterns2id->( 'apache', $-{version_number}[0] ) }
+					{custom} = 1;
 				next;
 			}
 			if ( $string =~ $L{apache_2} ) {
 				$self->note( 'grant(apache#2)', $-[0], $+[0] );
 				$gen_license->(
-					'apache', $1, $2,
-					$3 ? "bsd_${3}_clause" : ''
+					'apache', $+{version_number}, $+{version_later},
+					$+{version_bsd} ? "bsd_$+{version_bsd}_clause" : ''
 				);
-				$match{ $patterns2id->( 'apache', $1 ) }{custom} = 1;
+				$match{ $patterns2id->( 'apache', $+{version_number} ) }
+					{custom} = 1;
 				next;
 			}
 			if ( $string =~ $L{apache_4} ) {
 				$self->note( 'grant(apache#4)', $-[0], $+[0] );
-				$gen_license->( 'apache', $1, $2, 'mit', $3, $4 );
-				$match{ $patterns2id->( 'apache', $1 ) }{custom} = 1;
+				$gen_license->(
+					'apache', $+{version_number}, $+{version_later},
+					'mit',
+				);
+				$match{ $patterns2id->( 'apache', $+{version_number} ) }
+					{custom} = 1;
 				next;
 			}
 		}
@@ -825,12 +919,13 @@ method resolve ()
 	# FSFUL
 	# FIXME: add test covering this pattern
 	$log->trace('scan for FSFUL fulltext');
-	if (   !$license{fsful}
+	if (   !$fulltext{fsful}
 		and $string =~ $L{fsful} )
 	{
 		$self->tag(
 			Fulltext->new(
-				id    => [ "FSFUL~$1", "FSF Unlimited ($1 derivation)" ],
+				name  => "FSFUL~$1",
+				desc  => "FSF Unlimited ($1 derivation)",
 				begin => $-[0], end => $+[0],
 			)
 		);
@@ -840,15 +935,13 @@ method resolve ()
 	# FSFULLR
 	# FIXME: add test covering this pattern
 	$log->trace('scan for FSFULLR fulltext');
-	if (   !$license{fsfullr}
+	if (   !$fulltext{fsfullr}
 		and $string =~ $L{fsfullr} )
 	{
 		$self->tag(
 			Fulltext->new(
-				id => [
-					"FSFULLR~$1",
-					"FSF Unlimited (with Retention, $1 derivation)"
-				],
+				name  => "FSFULLR~$1",
+				desc  => "FSF Unlimited (with Retention, $1 derivation)",
 				begin => $-[0], end => $+[0],
 			)
 		);
@@ -857,7 +950,7 @@ method resolve ()
 
 	# usage
 	$log->trace('scan atomic for singleversion usage license grant');
-	foreach my $id (@L_type_usage) {
+	for my $id (@USAGE) {
 		next if ( $match{$id}{custom} );
 		if (    !$grant{$id}
 			and ( $L_grant_stepwise_incomplete{$id} or $force_atomic )
@@ -867,9 +960,12 @@ method resolve ()
 				$log->tracef( 'skip grant in covered range: %s', $id );
 			}
 			else {
-				$self->tag(
-					Grant->new( id => $id, begin => $-[0], end => $+[0] ) );
-				$grant{$id} = 1;
+				$grant{$id} = $self->tag(
+					Grant->new(
+						%{ $self->name_and_desc($id) },
+						begin => $-[0], end => $+[0]
+					)
+				);
 			}
 		}
 
@@ -893,8 +989,8 @@ method resolve ()
 
 	# singleversion
 	$log->trace('scan atomic for singleversion license grant');
-	foreach my $id (@L_type_singleversion) {
-		if (    !$license{$id}
+	for my $id (@SINGLEVERSION) {
+		if (    !$fulltext{$id}
 			and !$grant{$id}
 			and !$match{$id}{custom}
 			and ( $L_grant_stepwise_incomplete{$id} or $force_atomic )
@@ -904,13 +1000,16 @@ method resolve ()
 				$log->tracef( 'skip grant in covered range: %s', $id );
 			}
 			else {
-				$self->tag(
-					Grant->new( id => $id, begin => $-[0], end => $+[0] ) );
-				$grant{$id} = 1;
+				$grant{$id} = $self->tag(
+					Grant->new(
+						%{ $self->name_and_desc($id) },
+						begin => $-[0], end => $+[0]
+					)
+				);
 			}
 		}
 
-		if ( $license{$id} or $grant{$id} ) {
+		if ( $fulltext{$id} or $grant{$id} ) {
 			$gen_license->( $id2patterns->($id) )
 				unless ( $match{$id}{custom} );
 
@@ -924,13 +1023,11 @@ method resolve ()
 
 	# versioned
 	$log->trace('scan atomic for versioned license grant');
-	foreach my $id (@L_type_versioned) {
-		next if ( $match{$id}{custom} );
-
-		# skip embedded or referenced licenses
-		next if ( $license{rpsl_1} and grep { $id eq $_ } qw(mpl python) );
-
-		next if ( $license{$id} );
+	for my $id (@VERSIONED) {
+		next
+			if $match{$id}{custom}
+			or ( $fulltext{rpsl_1} and grep { $id eq $_ } qw(mpl python) )
+			or $fulltext{$id};
 		if (    !$grant{$id}
 			and ( $L_grant_stepwise_incomplete{$id} or $force_atomic )
 			and $RE{"GRANT_$id"}
@@ -940,37 +1037,35 @@ method resolve ()
 				$log->tracef( 'skip grant in covered range: %s', $id );
 			}
 			else {
-				$self->tag(
-					Grant->new( id => $id, begin => $-[0], end => $+[0] ) );
-				$grant{$id} = 1;
+				$grant{$id} = $self->tag(
+					Grant->new(
+						%{ $self->name_and_desc($id) },
+						begin => $-[0], end => $+[0]
+					)
+				);
 			}
 		}
-
 		if ( $grant{$id} ) {
 			$gen_license->($id);
 		}
 	}
 
 	# other
-	# TODO: add @L_type_group
+	# TODO: add @GROUP
 	$log->trace('scan atomic for misc fulltext/grant');
-	foreach my $id ( @L_type_unversioned, @L_type_combo ) {
-		next if ( !$license{$id} and $match{$id}{custom} );
-
+	for my $id ( @UNVERSIONED, @COMBO ) {
 		next
-			unless ( $license{$id}
-			or $grant{$id}
-			or $L_grant_stepwise_incomplete{$id}
-			or $force_atomic );
+			if (not $fulltext{$id}
+			and not $grant{$id}
+			and not $L_grant_stepwise_incomplete{$id}
+			and not $force_atomic )
+			or ( $fulltext{caldera}        and $id eq 'bsd' )
+			or ( $fulltext{cube}           and $id eq 'zlib' )
+			or ( $fulltext{dsdp}           and $id eq 'ntp' )
+			or ( $fulltext{mit_cmu}        and $id eq 'ntp_disclaimer' )
+			or ( $fulltext{ntp_disclaimer} and $id eq 'ntp' );
 
-		# skip embedded or referenced licenses
-		next if ( $license{caldera}        and $id eq 'bsd' );
-		next if ( $license{cube}           and $id eq 'zlib' );
-		next if ( $license{dsdp}           and $id eq 'ntp' );
-		next if ( $license{mit_cmu}        and $id eq 'ntp_disclaimer' );
-		next if ( $license{ntp_disclaimer} and $id eq 'ntp' );
-
-		if (    !$license{$id}
+		if (    !$fulltext{$id}
 			and !$grant{$id}
 			and $string =~ $RE{"GRANT_$id"} )
 		{
@@ -978,18 +1073,21 @@ method resolve ()
 				$log->tracef( 'skip grant in covered range: %s', $id );
 			}
 			else {
-				$self->tag(
-					Grant->new( id => $id, begin => $-[0], end => $+[0] ) );
-				$grant{$id} = 1;
+				$grant{$id} = $self->tag(
+					Grant->new(
+						%{ $self->name_and_desc($id) },
+						begin => $-[0], end => $+[0]
+					)
+				);
 			}
 		}
-		if ( $license{$id} or $grant{$id} ) {
+		if ( $fulltext{$id} or $grant{$id} ) {
 			$gen_license->($id);
 		}
 	}
 
 	# Expressions and exceptions contain DEP-5 or SPDX identifiers;
-	# flaws contains non-SPDX notes.
+	# flaws contain non-SPDX notes.
 	my ( $licenses, $exceptions, $flaws ) = $self->get_tags;
 
 	my @expressions = map { $_->[1] } @$licenses;
