@@ -12,7 +12,7 @@ String::License::Naming - base class for names of licenses and license naming sc
 
 =head1 VERSION
 
-Version v0.0.6
+Version v0.0.7
 
 =head1 DESCRIPTION
 
@@ -25,21 +25,15 @@ e.g. L<String::License::Naming::SPDX>.
 
 =cut
 
-package String::License::Naming v0.0.6;
+package String::License::Naming v0.0.7;
 
 use namespace::clean;
 
 class String::License::Naming;
 
-method list_schemes ()
-{
-	...;
-}
+method list_schemes () {...}
 
-method list_licenses ()
-{
-	...;
-}
+method list_licenses () {...}
 
 sub resolve_shortnames ( $keys, $schemes, $bootstrap=undef )
 {
